@@ -1,44 +1,35 @@
 import asyncio
-from datetime import datetime
+import datetime as dt
 
+from unittest import mock
 import pytest
 
 from bson import ObjectId
+import marshmallow as ma
+
+from pymongo.results import InsertOneResult, UpdateResult, DeleteResult
+from pymongo.collection import Collection
+from umongo import (
+    Document, EmbeddedDocument, MixinDocument, fields, exceptions, Reference
+)
+from umongo.document import MetaDocumentImplementation
+
+from .common import strip_indexes, name_sorted
+from ..common import BaseDBTest, TEST_DB
+
 
 DEP_ERROR = 'Missing motor'
 
 # Check if the required dependancies are met to run this driver's tests
 try:
     from motor.motor_asyncio import AsyncIOMotorClient
-    from motor import version_tuple as MOTOR_VERSION
 except ImportError:
     dep_error = True
 else:
     dep_error = False
 
-from ..common import BaseDBTest, TEST_DB
-
-from umongo import Document, EmbeddedDocument, fields, exceptions, Reference
-
 if not dep_error:  # Make sure the module is valid by importing it
     from umongo.frameworks import motor_asyncio as framework  # noqa
-    from pymongo.results import InsertOneResult, UpdateResult, DeleteResult
-
-
-def _stripped(indexes):
-    # With pymongo==2.8 a `ns` field is returned with Mongodb>=3 but
-    # not with MongoDB<2, thus we have to clean this before doing comparing
-    # Version may differ between database versions and configurations so it
-    # shall not be checked
-    return {
-        k: {sk: sv for sk, sv in v.items() if sk not in ('ns', 'v')}
-        for k, v in indexes.items()
-    }
-
-
-# Helper to sort indexes by name in order to have deterministic comparison
-def name_sorted(indexes):
-    return sorted(indexes, key=lambda x: x['name'])
 
 
 def make_db():
@@ -56,19 +47,19 @@ def loop():
 
 
 @pytest.mark.skipif(dep_error, reason=DEP_ERROR)
-class TestMotorAsyncio(BaseDBTest):
+class TestMotorAsyncIO(BaseDBTest):
 
     def test_create(self, loop, classroom_model):
         Student = classroom_model.Student
 
         async def do_test():
-            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+            john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
             ret = await john.commit()
             assert isinstance(ret, InsertOneResult)
             assert john.to_mongo() == {
                 '_id': john.id,
                 'name': 'John Doe',
-                'birthday': datetime(1995, 12, 12)
+                'birthday': dt.datetime(1995, 12, 12)
             }
 
             john2 = await Student.find_one(john.id)
@@ -83,7 +74,7 @@ class TestMotorAsyncio(BaseDBTest):
         Student = classroom_model.Student
 
         async def do_test():
-            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+            john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
             await john.commit()
             john.name = 'William Doe'
             assert john.to_mongo(update=True) == {'$set': {'name': 'William Doe'}}
@@ -103,17 +94,44 @@ class TestMotorAsyncio(BaseDBTest):
             await john.reload()
             assert john.name == 'Zorro Doe'
             # Cannot use conditions when creating document
-            with pytest.raises(RuntimeError):
+            with pytest.raises(exceptions.NotCreatedError):
                 await Student(name='Joe').commit(conditions={'name': 'dummy'})
 
         loop.run_until_complete(do_test())
 
+    def test_replace(self, loop, classroom_model):
+        Student = classroom_model.Student
+
+        async def do_test():
+            john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
+            # replace has no impact on creation
+            await john.commit(replace=True)
+            john.name = 'William Doe'
+            john.clear_modified()
+            ret = await john.commit(replace=True)
+            assert isinstance(ret, UpdateResult)
+            john2 = await Student.find_one(john.id)
+            assert john2._data == john._data
+            # Test conditional commit
+            john.name = 'Zorro Doe'
+            john.clear_modified()
+            with pytest.raises(exceptions.UpdateError):
+                await john.commit(conditions={'name': 'Bad Name'}, replace=True)
+            await john.commit(conditions={'name': 'William Doe'}, replace=True)
+            await john.reload()
+            assert john.name == 'Zorro Doe'
+            # Cannot use conditions when creating document
+            with pytest.raises(exceptions.NotCreatedError):
+                await Student(name='Joe').commit(conditions={'name': 'dummy'}, replace=True)
+
+        loop.run_until_complete(do_test())
+
     def test_remove(self, loop, classroom_model):
         Student = classroom_model.Student
 
         async def do_test():
             await Student.collection.drop()
-            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+            john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
             with pytest.raises(exceptions.NotCreatedError):
                 await john.remove()
             await john.commit()
@@ -145,7 +163,7 @@ class TestMotorAsyncio(BaseDBTest):
 
         async def do_test():
             await Student(name='Other dude').commit()
-            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+            john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
             with pytest.raises(exceptions.NotCreatedError):
                 await john.reload()
             await john.commit()
@@ -166,20 +184,11 @@ class TestMotorAsyncio(BaseDBTest):
             for i in range(10):
                 await Student(name='student-%s' % i).commit()
             cursor = Student.find(limit=5, skip=6)
-            if MOTOR_VERSION < (2, 0):
-                # this test doesn't make as much sense in motor>2.0, since the cursor no
-                # longer supports a count() method
-                count = await cursor.count()
-                assert count == 10
-                count_with_limit_and_skip = await cursor.count(with_limit_and_skip=True)
-                assert count_with_limit_and_skip == 4
-            else:
-                assert (await Student.count_documents()) == 10
-                assert (await Student.count_documents(limit=5, skip=6)) == 4
+            assert (await Student.count_documents()) == 10
+            assert (await Student.count_documents(limit=5, skip=6)) == 4
 
             # to_list with callback should fail
-            error = NotImplementedError if MOTOR_VERSION < (2, 0, 0) else TypeError
-            with pytest.raises(error):
+            with pytest.raises(TypeError):
                 await cursor.to_list(length=100, callback=lambda r, e: r if r else e)
 
             # Make sure returned documents are wrapped
@@ -246,7 +255,7 @@ class TestMotorAsyncio(BaseDBTest):
 
         async def do_test():
 
-            student = classroom_model.Student(name='Marty McFly', birthday=datetime(1968, 6, 9))
+            student = classroom_model.Student(name='Marty McFly', birthday=dt.datetime(1968, 6, 9))
             await student.commit()
             teacher = classroom_model.Teacher(name='M. Strickland')
             await teacher.commit()
@@ -260,7 +269,7 @@ class TestMotorAsyncio(BaseDBTest):
             assert student.to_mongo() == {
                 '_id': student.pk,
                 'name': 'Marty McFly',
-                'birthday': datetime(1968, 6, 9),
+                'birthday': dt.datetime(1968, 6, 9),
                 'courses': [course.pk]
             }
 
@@ -271,24 +280,24 @@ class TestMotorAsyncio(BaseDBTest):
         async def do_test():
 
             async def io_validate(field, value):
-                raise exceptions.ValidationError('Ho boys !')
+                raise ma.ValidationError('Ho boys !')
 
             @instance.register
             class Dummy(Document):
                 required_name = fields.StrField(required=True)
                 always_io_fail = fields.IntField(io_validate=io_validate)
 
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await Dummy().commit()
             assert exc.value.messages == {'required_name': ['Missing data for required field.']}
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await Dummy(required_name='required', always_io_fail=42).commit()
             assert exc.value.messages == {'always_io_fail': ['Ho boys !']}
 
             dummy = Dummy(required_name='required')
             await dummy.commit()
             del dummy.required_name
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await dummy.commit()
             assert exc.value.messages == {'required_name': ['Missing data for required field.']}
 
@@ -330,9 +339,14 @@ class TestMotorAsyncio(BaseDBTest):
             assert teacher_fetched.name == 'M. Strickland'
             # Test bad ref as well
             course.teacher = Reference(classroom_model.Teacher, ObjectId())
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await course.io_validate()
             assert exc.value.messages == {'teacher': ['Reference not found for document Teacher.']}
+            # Test setting to None / deleting
+            course.teacher = None
+            await course.io_validate()
+            del course.teacher
+            await course.io_validate()
 
         loop.run_until_complete(do_test())
 
@@ -370,7 +384,7 @@ class TestMotorAsyncio(BaseDBTest):
             Student = classroom_model.Student
 
             async def io_validate(field, value):
-                raise exceptions.ValidationError('Ho boys !')
+                raise ma.ValidationError('Ho boys !')
 
             @instance.register
             class EmbeddedDoc(EmbeddedDocument):
@@ -380,7 +394,12 @@ class TestMotorAsyncio(BaseDBTest):
             class IOStudent(Student):
                 io_field = fields.StrField(io_validate=io_validate)
                 list_io_field = fields.ListField(fields.IntField(io_validate=io_validate))
-                reference_io_field = fields.ReferenceField(classroom_model.Course, io_validate=io_validate)
+                dict_io_field = fields.DictField(
+                    fields.StrField(),
+                    fields.IntField(io_validate=io_validate),
+                )
+                reference_io_field = fields.ReferenceField(
+                    classroom_model.Course, io_validate=io_validate)
                 embedded_io_field = fields.EmbeddedField(EmbeddedDoc, io_validate=io_validate)
 
             bad_reference = ObjectId()
@@ -388,14 +407,16 @@ class TestMotorAsyncio(BaseDBTest):
                 name='Marty',
                 io_field='io?',
                 list_io_field=[1, 2],
+                dict_io_field={"1": 1, "2": 2},
                 reference_io_field=bad_reference,
                 embedded_io_field={'io_field': 42}
             )
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await student.io_validate()
             assert exc.value.messages == {
                 'io_field': ['Ho boys !'],
                 'list_io_field': {0: ['Ho boys !'], 1: ['Ho boys !']},
+                'dict_io_field': {"1": {"value": ['Ho boys !']}, "2": {"value": ['Ho boys !']}},
                 'reference_io_field': ['Ho boys !', 'Reference not found for document Course.'],
                 'embedded_io_field': {'io_field': ['Ho boys !']}
             }
@@ -464,12 +485,73 @@ class TestMotorAsyncio(BaseDBTest):
 
             @instance.register
             class IOStudent(Student):
-                io_field = fields.ListField(fields.IntField(io_validate=io_validate))
+                io_field = fields.ListField(
+                    fields.IntField(io_validate=io_validate),
+                    allow_none=True
+                )
 
             student = IOStudent(name='Marty', io_field=values)
             await student.io_validate()
             assert set(called) == set(values)
 
+            student.io_field = None
+            await student.io_validate()
+            del student.io_field
+            await student.io_validate()
+
+        loop.run_until_complete(do_test())
+
+    def test_io_validate_dict(self, loop, instance, classroom_model):
+
+        async def do_test():
+
+            Student = classroom_model.Student
+            called = []
+            keys = ["1", "2", "3", "4"]
+            values = [1, 2, 3, 4]
+
+            def io_validate(field, value):
+                called.append(value)
+
+            @instance.register
+            class IOStudent(Student):
+                io_field = fields.DictField(
+                    fields.StrField(),
+                    fields.IntField(io_validate=io_validate),
+                    allow_none=True
+                )
+
+            student = IOStudent(name='Marty', io_field=dict(zip(keys, values)))
+            await student.io_validate()
+            assert called == values
+
+            student.io_field = None
+            await student.io_validate()
+            del student.io_field
+            await student.io_validate()
+
+        loop.run_until_complete(do_test())
+
+    def test_io_validate_embedded(self, loop, instance, classroom_model):
+        Student = classroom_model.Student
+
+        @instance.register
+        class EmbeddedDoc(EmbeddedDocument):
+            io_field = fields.IntField()
+
+        @instance.register
+        class IOStudent(Student):
+            embedded_io_field = fields.EmbeddedField(EmbeddedDoc, allow_none=True)
+
+        async def do_test():
+
+            student = IOStudent(name='Marty', embedded_io_field={'io_field': 12})
+            await student.io_validate()
+            student.embedded_io_field = None
+            await student.io_validate()
+            del student.embedded_io_field
+            await student.io_validate()
+
         loop.run_until_complete(do_test())
 
     def test_indexes(self, loop, instance):
@@ -482,10 +564,9 @@ class TestMotorAsyncio(BaseDBTest):
                 no_indexed = fields.IntField()
 
                 class Meta:
-                    collection = 'simple_index_doc'
                     indexes = ['indexed']
 
-            await SimpleIndexDoc.collection.drop_indexes()
+            await SimpleIndexDoc.collection.drop()
 
             # Now ask for indexes building
             await SimpleIndexDoc.ensure_indexes()
@@ -498,12 +579,12 @@ class TestMotorAsyncio(BaseDBTest):
                     'key': [('indexed', 1)],
                 }
             }
-            assert _stripped(indexes) == expected_indexes
+            assert strip_indexes(indexes) == expected_indexes
 
             # Redoing indexes building should do nothing
             await SimpleIndexDoc.ensure_indexes()
             indexes = await SimpleIndexDoc.collection.index_information()
-            assert _stripped(indexes) == expected_indexes
+            assert strip_indexes(indexes) == expected_indexes
 
         loop.run_until_complete(do_test())
 
@@ -517,10 +598,9 @@ class TestMotorAsyncio(BaseDBTest):
                 no_indexed = fields.IntField()
 
                 class Meta:
-                    collection = 'simple_index_doc'
                     indexes = ['indexed']
 
-            await SimpleIndexDoc.collection.drop_indexes()
+            await SimpleIndexDoc.collection.drop()
 
             # Now ask for indexes building
             await SimpleIndexDoc.ensure_indexes()
@@ -533,12 +613,12 @@ class TestMotorAsyncio(BaseDBTest):
                     'key': [('indexed', 1)],
                 }
             }
-            assert _stripped(indexes) == expected_indexes
+            assert strip_indexes(indexes) == expected_indexes
 
             # Redoing indexes building should do nothing
             await SimpleIndexDoc.ensure_indexes()
             indexes = await SimpleIndexDoc.collection.index_information()
-            assert _stripped(indexes) == expected_indexes
+            assert strip_indexes(indexes) == expected_indexes
 
         loop.run_until_complete(do_test())
 
@@ -552,11 +632,7 @@ class TestMotorAsyncio(BaseDBTest):
                 sparse_unique = fields.IntField(unique=True)
                 required_unique = fields.IntField(unique=True, required=True)
 
-                class Meta:
-                    collection = 'unique_index_doc'
-
             await UniqueIndexDoc.collection.drop()
-            await UniqueIndexDoc.collection.drop_indexes()
 
             # Now ask for indexes building
             await UniqueIndexDoc.ensure_indexes()
@@ -575,19 +651,19 @@ class TestMotorAsyncio(BaseDBTest):
                     'sparse': True
                 }
             }
-            assert _stripped(indexes) == expected_indexes
+            assert strip_indexes(indexes) == expected_indexes
 
             # Redoing indexes building should do nothing
             await UniqueIndexDoc.ensure_indexes()
             indexes = await UniqueIndexDoc.collection.index_information()
-            assert _stripped(indexes) == expected_indexes
+            assert strip_indexes(indexes) == expected_indexes
 
             await UniqueIndexDoc(not_unique='a', required_unique=1).commit()
             await UniqueIndexDoc(not_unique='a', sparse_unique=1, required_unique=2).commit()
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await UniqueIndexDoc(not_unique='a', required_unique=1).commit()
             assert exc.value.messages == {'required_unique': 'Field value must be unique.'}
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await UniqueIndexDoc(not_unique='a', sparse_unique=1, required_unique=3).commit()
             assert exc.value.messages == {'sparse_unique': 'Field value must be unique.'}
 
@@ -604,18 +680,17 @@ class TestMotorAsyncio(BaseDBTest):
                 not_unique = fields.StrField()
 
                 class Meta:
-                    collection = 'unique_index_compound_doc'
                     # Must define custom index to do that
                     indexes = [{'key': ('compound1', 'compound2'), 'unique': True}]
 
             await UniqueIndexCompoundDoc.collection.drop()
-            await UniqueIndexCompoundDoc.collection.drop_indexes()
 
             # Now ask for indexes building
             await UniqueIndexCompoundDoc.ensure_indexes()
             indexes = await UniqueIndexCompoundDoc.collection.index_information()
             # Must sort compound indexes to avoid random inconsistence
-            indexes['compound1_1_compound2_1']['key'] = sorted(indexes['compound1_1_compound2_1']['key'])
+            indexes['compound1_1_compound2_1']['key'] = sorted(
+                indexes['compound1_1_compound2_1']['key'])
             expected_indexes = {
                 '_id_': {
                     'key': [('_id', 1)],
@@ -625,27 +700,28 @@ class TestMotorAsyncio(BaseDBTest):
                     'unique': True
                 }
             }
-            assert _stripped(indexes) == expected_indexes
+            assert strip_indexes(indexes) == expected_indexes
 
             # Redoing indexes building should do nothing
             await UniqueIndexCompoundDoc.ensure_indexes()
             indexes = await UniqueIndexCompoundDoc.collection.index_information()
             # Must sort compound indexes to avoid random inconsistence
-            indexes['compound1_1_compound2_1']['key'] = sorted(indexes['compound1_1_compound2_1']['key'])
-            assert _stripped(indexes) == expected_indexes
+            indexes['compound1_1_compound2_1']['key'] = sorted(
+                indexes['compound1_1_compound2_1']['key'])
+            assert strip_indexes(indexes) == expected_indexes
 
             # Index is on the tuple (compound1, compound2)
             await UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=1).commit()
             await UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=2).commit()
             await UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=1).commit()
             await UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=2).commit()
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=1).commit()
             assert exc.value.messages == {
                 'compound2': "Values of fields ['compound1', 'compound2'] must be unique together.",
                 'compound1': "Values of fields ['compound1', 'compound2'] must be unique together."
             }
-            with pytest.raises(exceptions.ValidationError) as exc:
+            with pytest.raises(ma.ValidationError) as exc:
                 await UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=1).commit()
             assert exc.value.messages == {
                 'compound2': "Values of fields ['compound1', 'compound2'] must be unique together.",
@@ -664,10 +740,6 @@ class TestMotorAsyncio(BaseDBTest):
                 not_unique = fields.StrField(unique=False)
                 unique = fields.IntField(unique=True)
 
-                class Meta:
-                    collection = 'unique_index_inheritance_doc'
-                    allow_inheritance = True
-
             @instance.register
             class UniqueIndexChildDoc(UniqueIndexParentDoc):
                 child_not_unique = fields.StrField(unique=False)
@@ -677,41 +749,34 @@ class TestMotorAsyncio(BaseDBTest):
                 class Meta:
                     indexes = ['manual_index']
 
-            await UniqueIndexChildDoc.collection.drop_indexes()
+            await UniqueIndexChildDoc.collection.drop()
 
             # Now ask for indexes building
             await UniqueIndexChildDoc.ensure_indexes()
-            indexes = []
-            async for index in UniqueIndexChildDoc.collection.list_indexes():
-                indexes.append(index)
+            indexes = [e async for e in UniqueIndexChildDoc.collection.list_indexes()]
             expected_indexes = [
                 {
                     'key': {'_id': 1},
                     'name': '_id_',
-                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB,
                 },
                 {
                     'key': {'unique': 1},
                     'name': 'unique_1',
                     'unique': True,
-                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                 },
                 {
                     'key': {'manual_index': 1, '_cls': 1},
                     'name': 'manual_index_1__cls_1',
-                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                 },
                 {
                     'key': {'_cls': 1},
                     'name': '_cls_1',
                     'unique': True,
-                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                 },
                 {
                     'key': {'child_unique': 1, '_cls': 1},
                     'name': 'child_unique_1__cls_1',
                     'unique': True,
-                    'ns': '%s.unique_index_inheritance_doc' % TEST_DB
                 }
             ]
             assert name_sorted(indexes) == name_sorted(expected_indexes)
@@ -731,17 +796,10 @@ class TestMotorAsyncio(BaseDBTest):
             class InheritanceSearchParent(Document):
                 pf = fields.IntField()
 
-                class Meta:
-                    collection = 'inheritance_search'
-                    allow_inheritance = True
-
             @instance.register
             class InheritanceSearchChild1(InheritanceSearchParent):
                 c1f = fields.IntField()
 
-                class Meta:
-                    allow_inheritance = True
-
             @instance.register
             class InheritanceSearchChild1Child(InheritanceSearchChild1):
                 sc1f = fields.IntField()
@@ -762,8 +820,8 @@ class TestMotorAsyncio(BaseDBTest):
             assert (await InheritanceSearchChild1Child.count_documents()) == 1
             assert (await InheritanceSearchChild2.count_documents()) == 1
 
-            res = await InheritanceSearchParent.find_one({'sc1f': 1})
-            assert isinstance(res, InheritanceSearchChild1Child)
+            res = await InheritanceSearchParent.find_one({'pf': 2})
+            assert isinstance(res, InheritanceSearchChild2)
 
             cursor = InheritanceSearchParent.find({'pf': 1})
             for r in (await cursor.to_list(length=100)):
@@ -835,11 +893,14 @@ class TestMotorAsyncio(BaseDBTest):
 
             res = await Book.count_documents({'title': 'The Hobbit'})
             assert res == 1
-            res = await Book.count_documents({'author.name': {'$in': ['JK Rowling', 'JRR Tolkien']}})
+            res = await Book.count_documents(
+                {'author.name': {'$in': ['JK Rowling', 'JRR Tolkien']}})
             assert res == 2
-            res = await Book.count_documents({'$and': [{'chapters.name': 'Roast Mutton'}, {'title': 'The Hobbit'}]})
+            res = await Book.count_documents(
+                {'$and': [{'chapters.name': 'Roast Mutton'}, {'title': 'The Hobbit'}]})
             assert res == 1
-            res = await Book.count_documents({'chapters.name': {'$all': ['Roast Mutton', 'A Short Rest']}})
+            res = await Book.count_documents(
+                {'chapters.name': {'$all': ['Roast Mutton', 'A Short Rest']}})
             assert res == 1
 
         loop.run_until_complete(do_test())
@@ -977,3 +1038,199 @@ class TestMotorAsyncio(BaseDBTest):
             await p.delete()
 
         loop.run_until_complete(do_test())
+
+    def test_mixin_pre_post_hooks(self, loop, instance):
+
+        async def do_test():
+
+            callbacks = []
+
+            @instance.register
+            class PrePostHooksMixin(MixinDocument):
+
+                def pre_insert(self):
+                    callbacks.append('pre_insert')
+
+                def pre_update(self):
+                    callbacks.append('pre_update')
+
+                def pre_delete(self):
+                    callbacks.append('pre_delete')
+
+                def post_insert(self, ret):
+                    assert isinstance(ret, InsertOneResult)
+                    callbacks.append('post_insert')
+
+                def post_update(self, ret):
+                    assert isinstance(ret, UpdateResult)
+                    callbacks.append('post_update')
+
+                def post_delete(self, ret):
+                    assert isinstance(ret, DeleteResult)
+                    callbacks.append('post_delete')
+
+            @instance.register
+            class Person(PrePostHooksMixin, Document):
+                name = fields.StrField()
+                age = fields.IntField()
+
+            p = Person(name='John', age=20)
+            await p.commit()
+            assert callbacks == ['pre_insert', 'post_insert']
+
+            callbacks.clear()
+            p.age = 22
+            await p.commit({'age': 22})
+            assert callbacks == ['pre_update', 'post_update']
+
+            callbacks.clear()
+            await p.delete()
+            assert callbacks == ['pre_delete', 'post_delete']
+
+        loop.run_until_complete(do_test())
+
+    def test_session_context_manager(self, loop, instance):
+        """Test session is passed to framework methods"""
+
+        coll_mock = mock.Mock(Collection, wraps=instance.db['Doc'])
+
+        class MockMetaDocumentImplementation(MetaDocumentImplementation):
+            @property
+            def collection(cls):
+                return coll_mock
+
+        @instance.register
+        class Doc(Document):
+            # Set unique to create an index
+            s = fields.StringField(unique=True)
+
+        Doc.__class__ = MockMetaDocumentImplementation
+
+        doc = Doc(s="test")
+
+        async def do_test():
+
+            async with instance.session() as session:
+                await doc.commit()
+            coll_mock.insert_one.assert_called_once()
+            assert coll_mock.insert_one.call_args[1]["session"] == session
+
+            doc.s = "retest"
+            async with instance.session() as session:
+                await doc.commit()
+            coll_mock.update_one.assert_called_once()
+            assert coll_mock.update_one.call_args[1]["session"] == session
+
+            doc.s = "reretest"
+            async with instance.session() as session:
+                await doc.commit(replace=True)
+            coll_mock.replace_one.assert_called_once()
+            assert coll_mock.replace_one.call_args[1]["session"] == session
+
+            async with instance.session() as session:
+                await doc.reload()
+            coll_mock.find_one.assert_called_once()
+            assert coll_mock.find_one.call_args[1]["session"] == session
+
+            async with instance.session() as session:
+                await doc.delete()
+            coll_mock.delete_one.assert_called_once()
+            assert coll_mock.delete_one.call_args[1]["session"] == session
+
+            coll_mock.reset_mock()
+            async with instance.session() as session:
+                await Doc.find_one(doc.id)
+            coll_mock.find_one.assert_called_once()
+            assert coll_mock.find_one.call_args[1]["session"] == session
+
+            async with instance.session() as session:
+                Doc.find()
+            coll_mock.find.assert_called_once()
+            assert coll_mock.find.call_args[1]["session"] == session
+
+            async with instance.session() as session:
+                await Doc.count_documents()
+            coll_mock.count_documents.assert_called_once()
+            assert coll_mock.count_documents.call_args[1]["session"] == session
+
+            async with instance.session() as session:
+                await Doc.ensure_indexes()
+            coll_mock.create_index.assert_called_once()
+            assert coll_mock.create_index.call_args[1]["session"] == session
+
+        loop.run_until_complete(do_test())
+
+    def test_2_to_3_migration(self, loop, db):
+
+        instance = framework.MotorAsyncIOMigrationInstance(db)
+
+        @instance.register
+        class AbstractEmbeddedDoc(EmbeddedDocument):
+            f = fields.StringField()
+
+            class Meta:
+                abstract = True
+
+        @instance.register
+        class ConcreteEmbeddedDoc(AbstractEmbeddedDoc):
+            pass
+
+        @instance.register
+        class ConcreteEmbeddedDocChild(ConcreteEmbeddedDoc):
+            pass
+
+        @instance.register
+        class AbstractDoc(Document):
+
+            class Meta:
+                abstract = True
+
+        @instance.register
+        class Doc(AbstractDoc):
+            ec = fields.EmbeddedField(ConcreteEmbeddedDoc)
+            ecc = fields.EmbeddedField(ConcreteEmbeddedDocChild)
+
+        @instance.register
+        class DocChild(Doc):
+            cec = fields.EmbeddedField(ConcreteEmbeddedDoc)
+            cecc = fields.EmbeddedField(ConcreteEmbeddedDocChild)
+
+        doc_umongo_2 = {
+            "ec": {"f": "Hello", "_cls": "ConcreteEmbeddedDoc"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+        child_doc_umongo_2 = {
+            "_cls": "DocChild",
+            "ec": {"f": "Hello", "_cls": "ConcreteEmbeddedDoc"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+            "cec": {"f": "Hello", "_cls": "ConcreteEmbeddedDoc"},
+            "cecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+
+        doc_umongo_3 = {
+            "ec": {"f": "Hello"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+        child_doc_umongo_3 = {
+            "_cls": "DocChild",
+            "ec": {"f": "Hello"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+            "cec": {"f": "Hello"},
+            "cecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+
+        async def do_test():
+
+            res = await instance.db.doc.insert_one(doc_umongo_2)
+            doc_umongo_3['_id'] = res.inserted_id
+            res = await instance.db.doc.insert_one(child_doc_umongo_2)
+            child_doc_umongo_3['_id'] = res.inserted_id
+
+            await instance.migrate_2_to_3()
+
+            res = await instance.db.doc.find_one(doc_umongo_3['_id'])
+            assert res == doc_umongo_3
+            res = await instance.db.doc.find_one(child_doc_umongo_3['_id'])
+            assert res == child_doc_umongo_3
+
+        loop.run_until_complete(do_test())
