@@ -1,12 +1,18 @@
-from datetime import datetime
+from functools import wraps
+import datetime as dt
 
 import pytest
 
 from bson import ObjectId
-from functools import wraps
+import marshmallow as ma
 
 from pymongo.results import InsertOneResult, UpdateResult, DeleteResult
 
+from umongo import (
+    Document, EmbeddedDocument, MixinDocument, fields, exceptions, Reference
+)
+
+from .common import strip_indexes, name_sorted
 from ..common import BaseDBTest, TEST_DB, con
 
 DEP_ERROR = 'Missing txmongo or pytest_twisted'
@@ -34,25 +40,11 @@ else:
     pytest_inlineCallbacks = pytest_twisted.inlineCallbacks
     dep_error = False
 
-from umongo import Document, EmbeddedDocument, fields, exceptions, Reference
-
 
 if not dep_error:  # Make sure the module is valid by importing it
     from umongo.frameworks import txmongo as framework  # noqa
 
 
-# Helper to sort indexes by name in order to have deterministic comparison
-def name_sorted(indexes):
-    return sorted(indexes, key=lambda x: x['name'])
-
-
-def _stripped(indexes):
-    # Version may differ between database versions and configurations so it shall not be checked
-    for idx in indexes:
-        idx.pop('v')
-    return indexes
-
-
 def make_db():
     return MongoConnection()[TEST_DB]
 
@@ -68,13 +60,13 @@ class TestTxMongo(BaseDBTest):
     @pytest_inlineCallbacks
     def test_create(self, classroom_model):
         Student = classroom_model.Student
-        john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+        john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
         ret = yield john.commit()
         assert isinstance(ret, InsertOneResult)
         assert john.to_mongo() == {
             '_id': john.id,
             'name': 'John Doe',
-            'birthday': datetime(1995, 12, 12)
+            'birthday': dt.datetime(1995, 12, 12)
         }
 
         john2 = yield Student.find_one(john.id)
@@ -86,7 +78,7 @@ class TestTxMongo(BaseDBTest):
     @pytest_inlineCallbacks
     def test_update(self, classroom_model):
         Student = classroom_model.Student
-        john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+        john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
         yield john.commit()
         john.name = 'William Doe'
         assert john.to_mongo(update=True) == {'$set': {'name': 'William Doe'}}
@@ -106,14 +98,38 @@ class TestTxMongo(BaseDBTest):
         yield john.reload()
         assert john.name == 'Zorro Doe'
         # Cannot use conditions when creating document
-        with pytest.raises(RuntimeError):
+        with pytest.raises(exceptions.NotCreatedError):
             yield Student(name='Joe').commit(conditions={'name': 'dummy'})
 
+    @pytest_inlineCallbacks
+    def test_replace(self, classroom_model):
+        Student = classroom_model.Student
+        john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
+        # replace has no impact on creation
+        yield john.commit(replace=True)
+        john.name = 'William Doe'
+        john.clear_modified()
+        ret = yield john.commit(replace=True)
+        assert isinstance(ret, UpdateResult)
+        john2 = yield Student.find_one(john.id)
+        assert john2._data == john._data
+        # Test conditional commit
+        john.name = 'Zorro Doe'
+        john.clear_modified()
+        with pytest.raises(exceptions.UpdateError):
+            yield john.commit(conditions={'name': 'Bad Name'}, replace=True)
+        yield john.commit(conditions={'name': 'William Doe'}, replace=True)
+        yield john.reload()
+        assert john.name == 'Zorro Doe'
+        # Cannot use conditions when creating document
+        with pytest.raises(exceptions.NotCreatedError):
+            yield Student(name='Joe').commit(conditions={'name': 'dummy'}, replace=True)
+
     @pytest_inlineCallbacks
     def test_delete(self, classroom_model):
         Student = classroom_model.Student
         yield Student.collection.drop()
-        john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+        john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
         with pytest.raises(exceptions.NotCreatedError):
             yield john.delete()
         yield john.commit()
@@ -145,7 +161,7 @@ class TestTxMongo(BaseDBTest):
     def test_reload(self, classroom_model):
         Student = classroom_model.Student
         yield Student(name='Other dude').commit()
-        john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
+        john = Student(name='John Doe', birthday=dt.datetime(1995, 12, 12))
         with pytest.raises(exceptions.NotCreatedError):
             yield john.reload()
         yield john.commit()
@@ -181,7 +197,7 @@ class TestTxMongo(BaseDBTest):
         Student.collection.drop()
         for i in range(10):
             yield Student(name='student-%s' % i).commit()
-        batch1, cursor1 = yield Student.find(limit=5, skip=6, cursor=True)
+        batch1, cursor1 = yield Student.find_with_cursor(limit=5, skip=6)
         assert len(batch1) == 4
         batch2, cursor2 = yield cursor1
         assert len(batch2) == 0
@@ -192,13 +208,13 @@ class TestTxMongo(BaseDBTest):
             names.append(elem.name)
         # Filter + projection
         assert sorted(names) == ['student-%s' % i for i in range(6, 10)]
-        batch1, cursor1 = yield Student.find({'name': 'student-0'}, ['name'], cursor=True)
+        batch1, cursor1 = yield Student.find_with_cursor({'name': 'student-0'}, ['name'])
         assert len(batch1) == 1
         assert batch1[0].name == 'student-0'
 
     @pytest_inlineCallbacks
     def test_classroom(self, classroom_model):
-        student = classroom_model.Student(name='Marty McFly', birthday=datetime(1968, 6, 9))
+        student = classroom_model.Student(name='Marty McFly', birthday=dt.datetime(1968, 6, 9))
         yield student.commit()
         teacher = classroom_model.Teacher(name='M. Strickland')
         yield teacher.commit()
@@ -212,7 +228,7 @@ class TestTxMongo(BaseDBTest):
         assert student.to_mongo() == {
             '_id': student.pk,
             'name': 'Marty McFly',
-            'birthday': datetime(1968, 6, 9),
+            'birthday': dt.datetime(1968, 6, 9),
             'courses': [course.pk]
         }
 
@@ -220,24 +236,24 @@ class TestTxMongo(BaseDBTest):
     def test_validation_on_commit(self, instance):
 
         def io_validate(field, value):
-            raise exceptions.ValidationError('Ho boys !')
+            raise ma.ValidationError('Ho boys !')
 
         @instance.register
         class Dummy(Document):
             required_name = fields.StrField(required=True)
             always_io_fail = fields.IntField(io_validate=io_validate)
 
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield Dummy().commit()
         assert exc.value.messages == {'required_name': ['Missing data for required field.']}
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield Dummy(required_name='required', always_io_fail=42).commit()
         assert exc.value.messages == {'always_io_fail': ['Ho boys !']}
 
         dummy = Dummy(required_name='required')
         yield dummy.commit()
         del dummy.required_name
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield dummy.commit()
         assert exc.value.messages == {'required_name': ['Missing data for required field.']}
 
@@ -275,9 +291,14 @@ class TestTxMongo(BaseDBTest):
         assert teacher_fetched.name == 'M. Strickland'
         # Test bad ref as well
         course.teacher = Reference(classroom_model.Teacher, ObjectId())
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield course.io_validate()
         assert exc.value.messages == {'teacher': ['Reference not found for document Teacher.']}
+        # Test setting to None / deleting
+        course.teacher = None
+        yield course.io_validate()
+        del course.teacher
+        yield course.io_validate()
 
     @pytest_inlineCallbacks
     def test_io_validate(self, instance, classroom_model):
@@ -308,7 +329,7 @@ class TestTxMongo(BaseDBTest):
         Student = classroom_model.Student
 
         def io_validate(field, value):
-            raise exceptions.ValidationError('Ho boys !')
+            raise ma.ValidationError('Ho boys !')
 
         @instance.register
         class EmbeddedDoc(EmbeddedDocument):
@@ -318,7 +339,12 @@ class TestTxMongo(BaseDBTest):
         class IOStudent(Student):
             io_field = fields.StrField(io_validate=io_validate)
             list_io_field = fields.ListField(fields.IntField(io_validate=io_validate))
-            reference_io_field = fields.ReferenceField(classroom_model.Course, io_validate=io_validate)
+            dict_io_field = fields.DictField(
+                fields.StrField(),
+                fields.IntField(io_validate=io_validate),
+            )
+            reference_io_field = fields.ReferenceField(
+                classroom_model.Course, io_validate=io_validate)
             embedded_io_field = fields.EmbeddedField(EmbeddedDoc, io_validate=io_validate)
 
         bad_reference = ObjectId()
@@ -326,14 +352,16 @@ class TestTxMongo(BaseDBTest):
             name='Marty',
             io_field='io?',
             list_io_field=[1, 2],
+            dict_io_field={"1": 1, "2": 2},
             reference_io_field=bad_reference,
             embedded_io_field={'io_field': 42}
         )
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield student.io_validate()
         assert exc.value.messages == {
             'io_field': ['Ho boys !'],
             'list_io_field': {0: ['Ho boys !'], 1: ['Ho boys !']},
+            'dict_io_field': {"1": {"value": ['Ho boys !']}, "2": {"value": ['Ho boys !']}},
             'reference_io_field': ['Ho boys !', 'Reference not found for document Course.'],
             'embedded_io_field': {'io_field': ['Ho boys !']}
         }
@@ -394,12 +422,63 @@ class TestTxMongo(BaseDBTest):
 
         @instance.register
         class IOStudent(Student):
-            io_field = fields.ListField(fields.IntField(io_validate=io_validate))
+            io_field = fields.ListField(fields.IntField(io_validate=io_validate), allow_none=True)
 
         student = IOStudent(name='Marty', io_field=values)
         yield student.io_validate()
         assert called == values
 
+        student.io_field = None
+        yield student.io_validate()
+        del student.io_field
+        yield student.io_validate()
+
+    @pytest_inlineCallbacks
+    def test_io_validate_dict(self, instance, classroom_model):
+        Student = classroom_model.Student
+        called = []
+        keys = ["1", "2", "3", "4"]
+        values = [1, 2, 3, 4]
+
+        def io_validate(field, value):
+            called.append(value)
+
+        @instance.register
+        class IOStudent(Student):
+            io_field = fields.DictField(
+                fields.StrField(),
+                fields.IntField(io_validate=io_validate),
+                allow_none=True
+            )
+
+        student = IOStudent(name='Marty', io_field=dict(zip(keys, values)))
+        yield student.io_validate()
+        assert called == values
+
+        student.io_field = None
+        yield student.io_validate()
+        del student.io_field
+        yield student.io_validate()
+
+    @pytest_inlineCallbacks
+    def test_io_validate_embedded(self, instance, classroom_model):
+        Student = classroom_model.Student
+
+        @instance.register
+        class EmbeddedDoc(EmbeddedDocument):
+            io_field = fields.IntField()
+
+        @instance.register
+        class IOStudent(Student):
+            embedded_io_field = fields.EmbeddedField(EmbeddedDoc, allow_none=True)
+
+        student = IOStudent(name='Marty', embedded_io_field={'io_field': 12})
+        yield student.io_validate()
+        student.embedded_io_field = None
+        yield student.io_validate()
+        del student.embedded_io_field
+        yield student.io_validate()
+
     @pytest_inlineCallbacks
     def test_indexes(self, instance):
 
@@ -409,10 +488,9 @@ class TestTxMongo(BaseDBTest):
             no_indexed = fields.IntField()
 
             class Meta:
-                collection = 'simple_index_doc'
                 indexes = ['indexed']
 
-        yield SimpleIndexDoc.collection.drop_indexes()
+        yield SimpleIndexDoc.collection.drop()
 
         # Now ask for indexes building
         yield SimpleIndexDoc.ensure_indexes()
@@ -422,20 +500,18 @@ class TestTxMongo(BaseDBTest):
             {
                 'key': {'_id': 1},
                 'name': '_id_',
-                'ns': '%s.simple_index_doc' % TEST_DB,
             },
             {
                 'key': {'indexed': 1},
                 'name': 'indexed_1',
-                'ns': '%s.simple_index_doc' % TEST_DB
             }
         ]
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
         # Redoing indexes building should do nothing
         yield SimpleIndexDoc.ensure_indexes()
         indexes = list(con[TEST_DB].simple_index_doc.list_indexes())
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
     @pytest_inlineCallbacks
     def test_indexes_inheritance(self, instance):
@@ -446,10 +522,9 @@ class TestTxMongo(BaseDBTest):
             no_indexed = fields.IntField()
 
             class Meta:
-                collection = 'simple_index_doc'
                 indexes = ['indexed']
 
-        yield SimpleIndexDoc.collection.drop_indexes()
+        yield SimpleIndexDoc.collection.drop()
 
         # Now ask for indexes building
         yield SimpleIndexDoc.ensure_indexes()
@@ -459,20 +534,18 @@ class TestTxMongo(BaseDBTest):
             {
                 'key': {'_id': 1},
                 'name': '_id_',
-                'ns': '%s.simple_index_doc' % TEST_DB,
             },
             {
                 'key': {'indexed': 1},
                 'name': 'indexed_1',
-                'ns': '%s.simple_index_doc' % TEST_DB
             }
         ]
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
         # Redoing indexes building should do nothing
         yield SimpleIndexDoc.ensure_indexes()
         indexes = list(con[TEST_DB].simple_index_doc.list_indexes())
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
     @pytest_inlineCallbacks
     def test_unique_index(self, instance):
@@ -483,11 +556,7 @@ class TestTxMongo(BaseDBTest):
             sparse_unique = fields.IntField(unique=True)
             required_unique = fields.IntField(unique=True, required=True)
 
-            class Meta:
-                collection = 'unique_index_doc'
-
         yield UniqueIndexDoc.collection.drop()
-        yield UniqueIndexDoc.collection.drop_indexes()
 
         # Now ask for indexes building
         yield UniqueIndexDoc.ensure_indexes()
@@ -496,35 +565,32 @@ class TestTxMongo(BaseDBTest):
             {
                 'key': {'_id': 1},
                 'name': '_id_',
-                'ns': '%s.unique_index_doc' % TEST_DB,
             },
             {
                 'key': {'required_unique': 1},
                 'name': 'required_unique_1',
                 'unique': True,
-                'ns': '%s.unique_index_doc' % TEST_DB
             },
             {
                 'key': {'sparse_unique': 1},
                 'name': 'sparse_unique_1',
                 'unique': True,
                 'sparse': True,
-                'ns': '%s.unique_index_doc' % TEST_DB
             },
         ]
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
         # Redoing indexes building should do nothing
         yield UniqueIndexDoc.ensure_indexes()
         indexes = list(con[TEST_DB].unique_index_doc.list_indexes())
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
         yield UniqueIndexDoc(not_unique='a', required_unique=1).commit()
         yield UniqueIndexDoc(not_unique='a', sparse_unique=1, required_unique=2).commit()
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield UniqueIndexDoc(not_unique='a', required_unique=1).commit()
         assert exc.value.messages == {'required_unique': 'Field value must be unique.'}
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield UniqueIndexDoc(not_unique='a', sparse_unique=1, required_unique=3).commit()
         assert exc.value.messages == {'sparse_unique': 'Field value must be unique.'}
 
@@ -538,12 +604,10 @@ class TestTxMongo(BaseDBTest):
             not_unique = fields.StrField()
 
             class Meta:
-                collection = 'unique_index_compound_doc'
                 # Must define custom index to do that
                 indexes = [{'key': ('compound1', 'compound2'), 'unique': True}]
 
         yield UniqueIndexCompoundDoc.collection.drop()
-        yield UniqueIndexCompoundDoc.collection.drop_indexes()
 
         # Now ask for indexes building
         yield UniqueIndexCompoundDoc.ensure_indexes()
@@ -552,34 +616,32 @@ class TestTxMongo(BaseDBTest):
             {
                 'key': {'_id': 1},
                 'name': '_id_',
-                'ns': '%s.unique_index_compound_doc' % TEST_DB,
             },
             {
                 'key': {'compound1': 1, 'compound2': 1},
                 'name': 'compound1_1_compound2_1',
                 'unique': True,
-                'ns': '%s.unique_index_compound_doc' % TEST_DB
             }
         ]
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
         # Redoing indexes building should do nothing
         yield UniqueIndexCompoundDoc.ensure_indexes()
         indexes = list(con[TEST_DB].unique_index_compound_doc.list_indexes())
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
         # Index is on the tuple (compound1, compound2)
         yield UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=1).commit()
         yield UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=2).commit()
         yield UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=1).commit()
         yield UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=2).commit()
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield UniqueIndexCompoundDoc(not_unique='a', compound1=1, compound2=1).commit()
         assert exc.value.messages == {
             'compound2': "Values of fields ['compound1', 'compound2'] must be unique together.",
             'compound1': "Values of fields ['compound1', 'compound2'] must be unique together."
         }
-        with pytest.raises(exceptions.ValidationError) as exc:
+        with pytest.raises(ma.ValidationError) as exc:
             yield UniqueIndexCompoundDoc(not_unique='a', compound1=2, compound2=1).commit()
         assert exc.value.messages == {
             'compound2': "Values of fields ['compound1', 'compound2'] must be unique together.",
@@ -595,10 +657,6 @@ class TestTxMongo(BaseDBTest):
             not_unique = fields.StrField(unique=False)
             unique = fields.IntField(unique=True)
 
-            class Meta:
-                collection = 'unique_index_inheritance_doc'
-                allow_inheritance = True
-
         @instance.register
         class UniqueIndexChildDoc(UniqueIndexParentDoc):
             child_not_unique = fields.StrField(unique=False)
@@ -608,7 +666,7 @@ class TestTxMongo(BaseDBTest):
             class Meta:
                 indexes = ['manual_index']
 
-        yield UniqueIndexChildDoc.collection.drop_indexes()
+        yield UniqueIndexChildDoc.collection.drop()
 
         # Now ask for indexes building
         yield UniqueIndexChildDoc.ensure_indexes()
@@ -617,7 +675,6 @@ class TestTxMongo(BaseDBTest):
             {
                 'key': {'_id': 1},
                 'name': '_id_',
-                'ns': '%s.unique_index_inheritance_doc' % TEST_DB,
                 'v': 1
             },
             {
@@ -625,35 +682,31 @@ class TestTxMongo(BaseDBTest):
                 'key': {'unique': 1},
                 'name': 'unique_1',
                 'unique': True,
-                'ns': '%s.unique_index_inheritance_doc' % TEST_DB
             },
             {
                 'v': 1,
                 'key': {'manual_index': 1, '_cls': 1},
                 'name': 'manual_index_1__cls_1',
-                'ns': '%s.unique_index_inheritance_doc' % TEST_DB
             },
             {
                 'v': 1,
                 'key': {'_cls': 1},
                 'name': '_cls_1',
                 'unique': True,
-                'ns': '%s.unique_index_inheritance_doc' % TEST_DB
             },
             {
                 'v': 1,
                 'key': {'child_unique': 1, '_cls': 1},
                 'name': 'child_unique_1__cls_1',
                 'unique': True,
-                'ns': '%s.unique_index_inheritance_doc' % TEST_DB
             }
         ]
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
         # Redoing indexes building should do nothing
         yield UniqueIndexChildDoc.ensure_indexes()
         indexes = list(con[TEST_DB].unique_index_inheritance_doc.list_indexes())
-        assert name_sorted(_stripped(indexes)) == name_sorted(expected_indexes)
+        assert name_sorted(strip_indexes(indexes)) == name_sorted(expected_indexes)
 
     @pytest_inlineCallbacks
     def test_inheritance_search(self, instance):
@@ -662,17 +715,10 @@ class TestTxMongo(BaseDBTest):
         class InheritanceSearchParent(Document):
             pf = fields.IntField()
 
-            class Meta:
-                collection = 'inheritance_search'
-                allow_inheritance = True
-
         @instance.register
         class InheritanceSearchChild1(InheritanceSearchParent):
             c1f = fields.IntField()
 
-            class Meta:
-                allow_inheritance = True
-
         @instance.register
         class InheritanceSearchChild1Child(InheritanceSearchChild1):
             sc1f = fields.IntField()
@@ -697,8 +743,8 @@ class TestTxMongo(BaseDBTest):
         res = yield InheritanceSearchChild2.find()
         assert len(res) == 1
 
-        res = yield InheritanceSearchParent.find_one({'sc1f': 1})
-        assert isinstance(res, InheritanceSearchChild1Child)
+        res = yield InheritanceSearchParent.find_one({'pf': 2})
+        assert isinstance(res, InheritanceSearchChild2)
 
         res = yield InheritanceSearchParent.find({'pf': 1})
         for r in res:
@@ -769,7 +815,8 @@ class TestTxMongo(BaseDBTest):
         assert len(res) == 1
         res = yield Book.find({'author.name': {'$in': ['JK Rowling', 'JRR Tolkien']}})
         assert len(res) == 2
-        res = yield Book.find({'$and': [{'chapters.name': 'Roast Mutton'}, {'title': 'The Hobbit'}]})
+        res = yield Book.find(
+            {'$and': [{'chapters.name': 'Roast Mutton'}, {'title': 'The Hobbit'}]})
         assert len(res) == 1
         res = yield Book.find({'chapters.name': {'$all': ['Roast Mutton', 'A Short Rest']}})
         assert len(res) == 1
@@ -896,3 +943,122 @@ class TestTxMongo(BaseDBTest):
         with pytest.raises(exceptions.DeleteError):
             yield p_concurrent.delete()
         yield p.delete()
+
+    @pytest_inlineCallbacks
+    def test_mixin_pre_post_hooks(self, instance):
+
+        callbacks = []
+
+        @instance.register
+        class PrePostHooksMixin(MixinDocument):
+
+            def pre_insert(self):
+                callbacks.append('pre_insert')
+
+            def pre_update(self):
+                callbacks.append('pre_update')
+
+            def pre_delete(self):
+                callbacks.append('pre_delete')
+
+            def post_insert(self, ret):
+                assert isinstance(ret, InsertOneResult)
+                callbacks.append('post_insert')
+
+            def post_update(self, ret):
+                assert isinstance(ret, UpdateResult)
+                callbacks.append('post_update')
+
+            def post_delete(self, ret):
+                assert isinstance(ret, DeleteResult)
+                callbacks.append('post_delete')
+
+        @instance.register
+        class Person(PrePostHooksMixin, Document):
+            name = fields.StrField()
+            age = fields.IntField()
+
+        p = Person(name='John', age=20)
+        yield p.commit()
+        assert callbacks == ['pre_insert', 'post_insert']
+
+        callbacks.clear()
+        p.age = 22
+        yield p.commit({'age': 22})
+        assert callbacks == ['pre_update', 'post_update']
+
+        callbacks.clear()
+        yield p.delete()
+        assert callbacks == ['pre_delete', 'post_delete']
+
+    @pytest_inlineCallbacks
+    def test_2_to_3_migration(self, db):
+
+        instance = framework.TxMongoMigrationInstance(db)
+
+        @instance.register
+        class AbstractEmbeddedDoc(EmbeddedDocument):
+            f = fields.StringField()
+
+            class Meta:
+                abstract = True
+
+        @instance.register
+        class ConcreteEmbeddedDoc(AbstractEmbeddedDoc):
+            pass
+
+        @instance.register
+        class ConcreteEmbeddedDocChild(ConcreteEmbeddedDoc):
+            pass
+
+        @instance.register
+        class AbstractDoc(Document):
+
+            class Meta:
+                abstract = True
+
+        @instance.register
+        class Doc(AbstractDoc):
+            ec = fields.EmbeddedField(ConcreteEmbeddedDoc)
+            ecc = fields.EmbeddedField(ConcreteEmbeddedDocChild)
+
+        @instance.register
+        class DocChild(Doc):
+            cec = fields.EmbeddedField(ConcreteEmbeddedDoc)
+            cecc = fields.EmbeddedField(ConcreteEmbeddedDocChild)
+
+        doc_umongo_2 = {
+            "ec": {"f": "Hello", "_cls": "ConcreteEmbeddedDoc"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+        child_doc_umongo_2 = {
+            "_cls": "DocChild",
+            "ec": {"f": "Hello", "_cls": "ConcreteEmbeddedDoc"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+            "cec": {"f": "Hello", "_cls": "ConcreteEmbeddedDoc"},
+            "cecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+
+        doc_umongo_3 = {
+            "ec": {"f": "Hello"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+        child_doc_umongo_3 = {
+            "_cls": "DocChild",
+            "ec": {"f": "Hello"},
+            "ecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+            "cec": {"f": "Hello"},
+            "cecc": {"f": "Hi", "_cls": "ConcreteEmbeddedDocChild"},
+        }
+
+        res = yield instance.db.doc.insert_one(doc_umongo_2)
+        doc_umongo_3['_id'] = res.inserted_id
+        res = yield instance.db.doc.insert_one(child_doc_umongo_2)
+        child_doc_umongo_3['_id'] = res.inserted_id
+
+        yield instance.migrate_2_to_3()
+
+        res = yield instance.db.doc.find_one(doc_umongo_3['_id'])
+        assert res == doc_umongo_3
+        res = yield instance.db.doc.find_one(child_doc_umongo_3['_id'])
+        assert res == child_doc_umongo_3
