@@ -1,19 +1,26 @@
+import collections
+from contextvars import ContextVar
+from contextlib import asynccontextmanager
+
+from inspect import iscoroutine
 import asyncio
 
 from motor.motor_asyncio import AsyncIOMotorDatabase, AsyncIOMotorCursor
-from motor import version_tuple as MOTOR_VERSION
 from pymongo.errors import DuplicateKeyError
-from inspect import iscoroutine
+import marshmallow as ma
 
 from ..builder import BaseBuilder
+from ..instance import Instance
 from ..document import DocumentImplementation
-from ..data_proxy import missing
 from ..data_objects import Reference
-from ..exceptions import NotCreatedError, UpdateError, ValidationError, DeleteError
-from ..fields import ReferenceField, ListField, EmbeddedField
+from ..exceptions import NotCreatedError, UpdateError, DeleteError, NoneReferenceError
+from ..fields import ReferenceField, ListField, DictField, EmbeddedField
 from ..query_mapper import map_query
 
-from .tools import cook_find_filter
+from .tools import cook_find_filter, remove_cls_field_from_embedded_docs
+
+
+SESSION = ContextVar("session", default=None)
 
 
 class WrappedCursor(AsyncIOMotorCursor):
@@ -22,7 +29,7 @@ class WrappedCursor(AsyncIOMotorCursor):
 
     def __init__(self, document_cls, cursor):
         # Such a cunning plan my lord !
-        # We inherit from Cursor but don't call it __init__ because
+        # We inherit from Cursor but don't call its __init__ because
         # we act as a proxy to the underlying raw_cursor
         WrappedCursor.raw_cursor.__set__(self, cursor)
         WrappedCursor.document_cls.__set__(self, document_cls)
@@ -120,13 +127,13 @@ class MotorAsyncIODocument(DocumentImplementation):
         """
         if not self.is_created:
             raise NotCreatedError("Document doesn't exists in database")
-        ret = await self.collection.find_one(self.pk)
+        ret = await self.collection.find_one(self.pk, session=SESSION.get())
         if ret is None:
             raise NotCreatedError("Document doesn't exists in database")
         self._data = self.DataProxy()
         self._data.from_mongo(ret)
 
-    async def commit(self, io_validate_all=False, conditions=None):
+    async def commit(self, io_validate_all=False, conditions=None, replace=False):
         """
         Commit the document in database.
         If the document doesn't already exist it will be inserted, otherwise
@@ -137,12 +144,13 @@ class MotorAsyncIODocument(DocumentImplementation):
             satisfies condition(s) (e.g. version number).
             Raises :class:`umongo.exceptions.UpdateError` if the
             conditions are not satisfied.
+        :param replace: Replace the document rather than update.
         :return: Update result dict returned by underlaying driver or
             ObjectId of the inserted document.
         """
         try:
             if self.is_created:
-                if self.is_modified():
+                if self.is_modified() or replace:
                     query = conditions or {}
                     query['_id'] = self.pk
                     # pre_update can provide additional query filter and/or
@@ -152,44 +160,48 @@ class MotorAsyncIODocument(DocumentImplementation):
                         query.update(map_query(additional_filter, self.schema.fields))
                     self.required_validate()
                     await self.io_validate(validate_all=io_validate_all)
-                    payload = self._data.to_mongo(update=True)
-                    ret = await self.collection.update_one(query, payload)
+                    if replace:
+                        payload = self._data.to_mongo(update=False)
+                        ret = await self.collection.replace_one(
+                            query, payload, session=SESSION.get())
+                    else:
+                        payload = self._data.to_mongo(update=True)
+                        ret = await self.collection.update_one(
+                            query, payload, session=SESSION.get())
                     if ret.matched_count != 1:
                         raise UpdateError(ret)
                     await self.__coroutined_post_update(ret)
                 else:
                     ret = None
             elif conditions:
-                raise RuntimeError('Document must already exist in database to use `conditions`.')
+                raise NotCreatedError(
+                    'Document must already exist in database to use `conditions`.'
+                )
             else:
                 await self.__coroutined_pre_insert()
                 self.required_validate()
                 await self.io_validate(validate_all=io_validate_all)
                 payload = self._data.to_mongo(update=False)
-                ret = await self.collection.insert_one(payload)
+                ret = await self.collection.insert_one(payload, session=SESSION.get())
                 # TODO: check ret ?
-                self._data.set_by_mongo_name('_id', ret.inserted_id)
+                self._data.set(self.pk_field, ret.inserted_id)
                 self.is_created = True
                 await self.__coroutined_post_insert(ret)
         except DuplicateKeyError as exc:
-            # Need to dig into error message to find faulting index
-            errmsg = exc.details['errmsg']
-            for index in self.opts.indexes:
-                if ('.$%s' % index.document['name'] in errmsg or
-                        ' %s ' % index.document['name'] in errmsg):
-                    keys = index.document['key'].keys()
-                    if len(keys) == 1:
-                        key = tuple(keys)[0]
-                        msg = self.schema.fields[key].error_messages['unique']
-                        raise ValidationError({key: msg})
-                    else:
-                        fields = self.schema.fields
-                        # Compound index (sort value to make testing easier)
-                        keys = sorted(keys)
-                        raise ValidationError({k: fields[k].error_messages[
-                            'unique_compound'].format(fields=keys) for k in keys})
-            # Unknown index, cannot wrap the error so just reraise it
-            raise
+            # Sort value to make testing easier for compound indexes
+            keys = sorted(exc.details['keyPattern'].keys())
+            try:
+                fields = [self.schema.fields[k] for k in keys]
+            except KeyError:
+                # A key in the index is unknwon from umongo
+                raise exc
+            if len(keys) == 1:
+                msg = fields[0].error_messages['unique']
+                raise ma.ValidationError({keys[0]: msg})
+            raise ma.ValidationError({
+                k: f.error_messages['unique_compound'].format(fields=keys)
+                for k, f in zip(keys, fields)
+            })
         self._data.clear_modified()
         return ret
 
@@ -222,7 +234,7 @@ class MotorAsyncIODocument(DocumentImplementation):
         additional_filter = await self.__coroutined_pre_delete()
         if additional_filter:
             query.update(map_query(additional_filter, self.schema.fields))
-        ret = await self.collection.delete_one(query)
+        ret = await self.collection.delete_one(query, session=SESSION.get())
         if ret.deleted_count != 1:
             raise DeleteError(ret)
         self.is_created = False
@@ -238,9 +250,8 @@ class MotorAsyncIODocument(DocumentImplementation):
         """
         if validate_all:
             return await _io_validate_data_proxy(self.schema, self._data)
-        else:
-            return await _io_validate_data_proxy(
-                self.schema, self._data, partial=self._data.get_modified_fields())
+        return await _io_validate_data_proxy(
+            self.schema, self._data, partial=self._data.get_modified_fields())
 
     @classmethod
     async def find_one(cls, filter=None, *args, **kwargs):
@@ -248,7 +259,7 @@ class MotorAsyncIODocument(DocumentImplementation):
         Find a single document in database.
         """
         filter = cook_find_filter(cls, filter)
-        ret = await cls.collection.find_one(filter, *args, **kwargs)
+        ret = await cls.collection.find_one(filter, session=SESSION.get(), *args, **kwargs)
         if ret is not None:
             ret = cls.build_from_mongo(ret, use_cls=True)
         return ret
@@ -261,7 +272,10 @@ class MotorAsyncIODocument(DocumentImplementation):
         Returns a cursor that provide Documents.
         """
         filter = cook_find_filter(cls, filter)
-        return WrappedCursor(cls, cls.collection.find(filter, *args, **kwargs))
+        return WrappedCursor(
+            cls,
+            cls.collection.find(filter, session=SESSION.get(), *args, **kwargs)
+        )
 
     @classmethod
     async def count_documents(cls, filter=None, *, with_limit_and_skip=False, **kwargs):
@@ -269,21 +283,17 @@ class MotorAsyncIODocument(DocumentImplementation):
         Return a count of the documents in a collection.
         """
         filter = cook_find_filter(cls, filter or {})
-        if MOTOR_VERSION < (2, 0):
-            count = await cls.find(filter, **kwargs).count(with_limit_and_skip)
-            return count
-        count = await cls.collection.count_documents(filter, **kwargs)
-        return count
+        return await cls.collection.count_documents(filter, session=SESSION.get(), **kwargs)
 
     @classmethod
     async def ensure_indexes(cls):
         """
         Check&create if needed the Document's indexes in database
         """
-        for index in cls.opts.indexes:
+        for index in cls.indexes:
             kwargs = index.document.copy()
-            keys = [(k, d) for k, d in kwargs.pop('key').items()]
-            await cls.collection.create_index(keys, **kwargs)
+            keys = kwargs.pop('key').items()
+            await cls.collection.create_index(keys, session=SESSION.get(), **kwargs)
 
 
 # Run multiple validators and collect all errors in one
@@ -292,12 +302,12 @@ async def _run_validators(validators, field, value):
     tasks = [validator(field, value) for validator in validators]
     results = await asyncio.gather(*tasks, return_exceptions=True)
     for i, res in enumerate(results):
-        if isinstance(res, ValidationError):
+        if isinstance(res, ma.ValidationError):
             errors.extend(res.messages)
         elif res:
             raise res
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
 
 
 async def _io_validate_data_proxy(schema, data_proxy, partial=None):
@@ -307,9 +317,8 @@ async def _io_validate_data_proxy(schema, data_proxy, partial=None):
     for name, field in schema.fields.items():
         if partial and name not in partial:
             continue
-        data_name = field.attribute or name
-        value = data_proxy._data[data_name]
-        if value is missing:
+        value = data_proxy.get(name)
+        if value is ma.missing:
             continue
         try:
             if field.io_validate_recursive:
@@ -317,39 +326,65 @@ async def _io_validate_data_proxy(schema, data_proxy, partial=None):
             if field.io_validate:
                 tasks.append(_run_validators(field.io_validate, field, value))
                 tasks_field_name.append(name)
-        except ValidationError as ve:
-            errors[name] = ve.messages
+        except ma.ValidationError as exc:
+            errors[name] = exc.messages
     results = await asyncio.gather(*tasks, return_exceptions=True)
     for i, res in enumerate(results):
-        if isinstance(res, ValidationError):
+        if isinstance(res, ma.ValidationError):
             errors[tasks_field_name[i]] = res.messages
         elif res:
             raise res
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
 
 
 async def _reference_io_validate(field, value):
+    if value is None:
+        return
     await value.fetch(no_data=True)
 
 
 async def _list_io_validate(field, value):
-    validators = field.container.io_validate
-    if not validators or not value:
+    if not value:
         return
-    tasks = [_run_validators(validators, field.container, e) for e in value]
+    validators = field.inner.io_validate
+    if not validators:
+        return
+    tasks = [_run_validators(validators, field.inner, e) for e in value]
     results = await asyncio.gather(*tasks, return_exceptions=True)
     errors = {}
     for i, res in enumerate(results):
-        if isinstance(res, ValidationError):
+        if isinstance(res, ma.ValidationError):
             errors[i] = res.messages
         elif res:
             raise res
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
+
+
+async def _dict_io_validate(field, value):
+    if not value or not field.value_field:
+        return
+    validators = field.value_field.io_validate
+    if not validators:
+        return
+    tasks = []
+    for key, val in value.items():
+        tasks.append(_run_validators(validators, field.value_field, val))
+    results = await asyncio.gather(*tasks, return_exceptions=True)
+    errors = collections.defaultdict(dict)
+    for key, res in zip(value.keys(), results):
+        if isinstance(res, ma.ValidationError):
+            errors[key]["value"] = res.messages
+        elif res:
+            raise res
+    if errors:
+        raise ma.ValidationError(errors)
 
 
 async def _embedded_document_io_validate(field, value):
+    if not value:
+        return
     await _io_validate_data_proxy(value.schema, value._data)
 
 
@@ -362,10 +397,10 @@ class MotorAsyncIOReference(Reference):
     async def fetch(self, no_data=False, force_reload=False):
         if not self._document or force_reload:
             if self.pk is None:
-                raise ReferenceError('Cannot retrieve a None Reference')
+                raise NoneReferenceError('Cannot retrieve a None Reference')
             self._document = await self.document_cls.find_one(self.pk)
             if not self._document:
-                raise ValidationError(self.error_messages['not_found'].format(
+                raise ma.ValidationError(self.error_messages['not_found'].format(
                     document=self.document_cls.__name__))
         return self._document
 
@@ -374,10 +409,6 @@ class MotorAsyncIOBuilder(BaseBuilder):
 
     BASE_DOCUMENT_CLS = MotorAsyncIODocument
 
-    @staticmethod
-    def is_compatible_with(db):
-        return isinstance(db, AsyncIOMotorDatabase)
-
     def _patch_field(self, field):
         super()._patch_field(field)
 
@@ -389,12 +420,61 @@ class MotorAsyncIOBuilder(BaseBuilder):
                 validators = list(validators)
             else:
                 validators = [validators]
-            field.io_validate = [v if asyncio.iscoroutinefunction(v) else asyncio.coroutine(v)
-                                 for v in validators]
+            field.io_validate = [
+                v if asyncio.iscoroutinefunction(v) else asyncio.coroutine(v)
+                for v in validators
+            ]
         if isinstance(field, ListField):
             field.io_validate_recursive = _list_io_validate
+        if isinstance(field, DictField):
+            field.io_validate_recursive = _dict_io_validate
         if isinstance(field, ReferenceField):
             field.io_validate.append(_reference_io_validate)
             field.reference_cls = MotorAsyncIOReference
         if isinstance(field, EmbeddedField):
             field.io_validate_recursive = _embedded_document_io_validate
+
+
+class MotorAsyncIOInstance(Instance):
+    """
+    :class:`umongo.instance.Instance` implementation for motor-asyncio
+    """
+    BUILDER_CLS = MotorAsyncIOBuilder
+
+    @staticmethod
+    def is_compatible_with(db):
+        return isinstance(db, AsyncIOMotorDatabase)
+
+    @asynccontextmanager
+    async def session(self):
+        async with await self.db.client.start_session() as session:
+            try:
+                token = SESSION.set(session)
+                yield session
+            finally:
+                SESSION.reset(token)
+
+
+class MotorAsyncIOMigrationInstance(MotorAsyncIOInstance):
+    """AsyncIO instance with migration features"""
+
+    async def migrate_2_to_3(self):
+        """Migrate database from umongo 2 to umongo 3
+
+        - EmbeddedDocument _cls field is only set if child of concrete embedded document
+        """
+        concrete_not_children = [
+            name for name, ed in self._embedded_lookup.items()
+            if not ed.opts.is_child and not ed.opts.abstract
+        ]
+
+        for doc_cls in self._doc_lookup.values():
+            if doc_cls.opts.abstract:
+                continue
+            if doc_cls.opts.is_child:
+                continue
+            async for doc in doc_cls.collection.find():
+                doc = remove_cls_field_from_embedded_docs(doc, concrete_not_children)
+                ret = await doc_cls.collection.replace_one({"_id": doc["_id"]}, doc)
+                if ret.matched_count != 1:
+                    raise UpdateError(ret)
