@@ -1,18 +1,19 @@
-from txmongo.database import Database
 from twisted.internet.defer import (
     inlineCallbacks, Deferred, DeferredList, returnValue, maybeDeferred)
 from txmongo import filter as qf
+from txmongo.database import Database
 from pymongo.errors import DuplicateKeyError
+import marshmallow as ma
 
 from ..builder import BaseBuilder
+from ..instance import Instance
 from ..document import DocumentImplementation
-from ..data_proxy import missing
 from ..data_objects import Reference
-from ..exceptions import NotCreatedError, UpdateError, DeleteError, ValidationError
-from ..fields import ReferenceField, ListField, EmbeddedField
+from ..exceptions import NotCreatedError, UpdateError, DeleteError, NoneReferenceError
+from ..fields import ReferenceField, ListField, DictField, EmbeddedField
 from ..query_mapper import map_query
 
-from .tools import cook_find_filter
+from .tools import cook_find_filter, remove_cls_field_from_embedded_docs
 
 
 class TxMongoDocument(DocumentImplementation):
@@ -38,7 +39,7 @@ class TxMongoDocument(DocumentImplementation):
         self._data.from_mongo(ret)
 
     @inlineCallbacks
-    def commit(self, io_validate_all=False, conditions=None):
+    def commit(self, io_validate_all=False, conditions=None, replace=False):
         """
         Commit the document in database.
         If the document doesn't already exist it will be inserted, otherwise
@@ -49,12 +50,13 @@ class TxMongoDocument(DocumentImplementation):
             satisfies condition(s) (e.g. version number).
             Raises :class:`umongo.exceptions.UpdateError` if the
             conditions are not satisfied.
-         :return: A :class:`pymongo.results.UpdateResult` or
+        :param replace: Replace the document rather than update.
+        :return: A :class:`pymongo.results.UpdateResult` or
             :class:`pymongo.results.InsertOneResult` depending of the operation.
        """
         try:
             if self.is_created:
-                if self.is_modified():
+                if self.is_modified() or replace:
                     query = conditions or {}
                     query['_id'] = self.pk
                     # pre_update can provide additional query filter and/or
@@ -64,15 +66,21 @@ class TxMongoDocument(DocumentImplementation):
                         query.update(map_query(additional_filter, self.schema.fields))
                     self.required_validate()
                     yield self.io_validate(validate_all=io_validate_all)
-                    payload = self._data.to_mongo(update=True)
-                    ret = yield self.collection.update_one(query, payload)
+                    if replace:
+                        payload = self._data.to_mongo(update=False)
+                        ret = yield self.collection.replace_one(query, payload)
+                    else:
+                        payload = self._data.to_mongo(update=True)
+                        ret = yield self.collection.update_one(query, payload)
                     if ret.matched_count != 1:
                         raise UpdateError(ret)
                     yield maybeDeferred(self.post_update, ret)
                 else:
                     ret = None
             elif conditions:
-                raise RuntimeError('Document must already exist in database to use `conditions`.')
+                raise NotCreatedError(
+                    'Document must already exist in database to use `conditions`.'
+                )
             else:
                 yield maybeDeferred(self.pre_insert)
                 self.required_validate()
@@ -80,25 +88,24 @@ class TxMongoDocument(DocumentImplementation):
                 payload = self._data.to_mongo(update=False)
                 ret = yield self.collection.insert_one(payload)
                 # TODO: check ret ?
-                self._data.set_by_mongo_name('_id', ret.inserted_id)
+                self._data.set(self.pk_field, ret.inserted_id)
                 self.is_created = True
                 yield maybeDeferred(self.post_insert, ret)
         except DuplicateKeyError as exc:
-            # Need to dig into error message to find faulting index
-            errmsg = exc.details['errmsg']
-            for index in self.opts.indexes:
-                if ('.$%s' % index.document['name'] in errmsg or
-                        ' %s ' % index.document['name'] in errmsg):
-                    keys = index.document['key'].keys()
-                    if len(keys) == 1:
-                        msg = self.schema.fields[keys[0]].error_messages['unique']
-                        raise ValidationError({keys[0]: msg})
-                    else:
-                        fields = self.schema.fields
-                        # Compound index (sort value to make testing easier)
-                        keys = sorted(keys)
-                        raise ValidationError({k: fields[k].error_messages[
-                            'unique_compound'].format(fields=keys) for k in keys})
+            # Sort value to make testing easier for compound indexes
+            keys = sorted(exc.details['keyPattern'].keys())
+            try:
+                fields = [self.schema.fields[k] for k in keys]
+            except KeyError:
+                # A key in the index is unknwon from umongo
+                raise exc
+            if len(keys) == 1:
+                msg = fields[0].error_messages['unique']
+                raise ma.ValidationError({keys[0]: msg})
+            raise ma.ValidationError({
+                k: f.error_messages['unique_compound'].format(fields=keys)
+                for k, f in zip(keys, fields)
+            })
         self._data.clear_modified()
         return ret
 
@@ -142,54 +149,59 @@ class TxMongoDocument(DocumentImplementation):
         """
         if validate_all:
             return _io_validate_data_proxy(self.schema, self._data)
-        else:
-            return _io_validate_data_proxy(
-                self.schema, self._data, partial=self._data.get_modified_fields())
+        return _io_validate_data_proxy(
+            self.schema, self._data, partial=self._data.get_modified_fields())
 
     @classmethod
     @inlineCallbacks
-    def find_one(cls, spec=None, *args, **kwargs):
+    def find_one(cls, filter=None, *args, **kwargs):
         """
         Find a single document in database.
         """
-        # In txmongo, `spec` is for filtering and `filter` is for sorting
-        spec = cook_find_filter(cls, spec)
-        ret = yield cls.collection.find_one(spec, *args, **kwargs)
+        filter = cook_find_filter(cls, filter)
+        ret = yield cls.collection.find_one(filter, *args, **kwargs)
         if ret is not None:
             ret = cls.build_from_mongo(ret, use_cls=True)
         return ret
 
     @classmethod
     @inlineCallbacks
-    def find(cls, spec=None, *args, **kwargs):
+    def find(cls, filter=None, *args, **kwargs):
         """
         Find a list document in database.
 
-        Returns a cursor that provide Documents.
+        Returns a list of Documents.
         """
-        # In txmongo, `spec` is for filtering and `filter` is for sorting
-        spec = cook_find_filter(cls, spec)
-        raw_cursor_or_list = yield cls.collection.find(spec, *args, **kwargs)
-        if isinstance(raw_cursor_or_list, tuple):
-
-            def wrap_raw_results(result):
-                cursor = result[1]
-                if cursor is not None:
-                    cursor.addCallback(wrap_raw_results)
-                return ([cls.build_from_mongo(e, use_cls=True) for e in result[0]], cursor)
-
-            return wrap_raw_results(raw_cursor_or_list)
-        else:
-            return [cls.build_from_mongo(e, use_cls=True) for e in raw_cursor_or_list]
+        filter = cook_find_filter(cls, filter)
+        raw_cursor_or_list = yield cls.collection.find(filter, *args, **kwargs)
+        return [cls.build_from_mongo(e, use_cls=True) for e in raw_cursor_or_list]
 
     @classmethod
-    def count(cls, spec=None, **kwargs):
+    @inlineCallbacks
+    def find_with_cursor(cls, filter=None, *args, **kwargs):
+        """
+        Find a list document in database.
+
+        Returns a cursor that provides Documents.
+        """
+        filter = cook_find_filter(cls, filter)
+        raw_cursor_or_list = yield cls.collection.find_with_cursor(filter, *args, **kwargs)
+
+        def wrap_raw_results(result):
+            cursor = result[1]
+            if cursor is not None:
+                cursor.addCallback(wrap_raw_results)
+            return ([cls.build_from_mongo(e, use_cls=True) for e in result[0]], cursor)
+
+        return wrap_raw_results(raw_cursor_or_list)
+
+    @classmethod
+    def count(cls, filter=None, **kwargs):
         """
         Get the number of documents in this collection.
         """
-        # In txmongo, `spec` is for filtering and `filter` is for sorting
-        spec = cook_find_filter(cls, spec)
-        return cls.collection.count(spec=spec, **kwargs)
+        filter = cook_find_filter(cls, filter)
+        return cls.collection.count(filter=filter, **kwargs)
 
     @classmethod
     @inlineCallbacks
@@ -197,21 +209,24 @@ class TxMongoDocument(DocumentImplementation):
         """
         Check&create if needed the Document's indexes in database
         """
-        for index in cls.opts.indexes:
+        for index in cls.indexes:
             kwargs = index.document.copy()
             keys = kwargs.pop('key')
-            index = qf.sort([(k, d) for k, d in keys.items()])
+            index = qf.sort(keys.items())
             yield cls.collection.create_index(index, **kwargs)
 
 
-def _errback_factory(errors, field=None):
+def _errback_factory(errors, field=None, subkey=None):
 
     def errback(err):
-        if isinstance(err.value, ValidationError):
+        if isinstance(err.value, ma.ValidationError):
+            error = err.value.messages
+            if subkey is not None:
+                error = {subkey: error}
             if field is not None:
-                errors[field] = err.value.messages
+                errors[field] = error
             else:
-                errors.extend(err.value.messages)
+                errors.extend(error)
         else:
             raise err.value
 
@@ -226,15 +241,17 @@ def _run_validators(validators, field, value):
     for validator in validators:
         try:
             defer = validator(field, value)
-        except ValidationError as ve:
-            errors.extend(ve.messages)
+        except ma.ValidationError as exc:
+            errors.extend(exc.messages)
         else:
+            if defer is None:
+                continue
             assert isinstance(defer, Deferred), 'io_validate functions must return a Deferred'
             defer.addErrback(_errback_factory(errors))
             defers.append(defer)
     yield DeferredList(defers)
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
 
 
 @inlineCallbacks
@@ -244,9 +261,8 @@ def _io_validate_data_proxy(schema, data_proxy, partial=None):
     for name, field in schema.fields.items():
         if partial and name not in partial:
             continue
-        data_name = field.attribute or name
-        value = data_proxy._data[data_name]
-        if value is missing:
+        value = data_proxy.get(name)
+        if value is ma.missing:
             continue
         try:
             if field.io_validate_recursive:
@@ -255,34 +271,59 @@ def _io_validate_data_proxy(schema, data_proxy, partial=None):
                 defer = _run_validators(field.io_validate, field, value)
                 defer.addErrback(_errback_factory(errors, name))
                 defers.append(defer)
-        except ValidationError as ve:
-            errors[name] = ve.messages
+        except ma.ValidationError as exc:
+            errors[name] = exc.messages
     yield DeferredList(defers)
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
 
 
+@inlineCallbacks
 def _reference_io_validate(field, value):
-    return value.fetch(no_data=True)
+    if value is None:
+        return
+    yield value.fetch(no_data=True)
 
 
 @inlineCallbacks
 def _list_io_validate(field, value):
-    validators = field.container.io_validate
-    if not validators or not value:
+    if not value:
+        return
+    validators = field.inner.io_validate
+    if not validators:
         return
     errors = {}
     defers = []
-    for i, e in enumerate(value):
-        defer = _run_validators(validators, field.container, e)
-        defer.addErrback(_errback_factory(errors, i))
+    for idx, exc in enumerate(value):
+        defer = _run_validators(validators, field.inner, exc)
+        defer.addErrback(_errback_factory(errors, idx))
         defers.append(defer)
     yield DeferredList(defers)
     if errors:
-        raise ValidationError(errors)
+        raise ma.ValidationError(errors)
+
+
+@inlineCallbacks
+def _dict_io_validate(field, value):
+    if not value or not field.value_field:
+        return
+    validators = field.value_field.io_validate
+    if not validators:
+        return
+    errors = {}
+    defers = []
+    for key, exc in value.items():
+        defer = _run_validators(validators, field.value_field, exc)
+        defer.addErrback(_errback_factory(errors, key, subkey="value"))
+        defers.append(defer)
+    yield DeferredList(defers)
+    if errors:
+        raise ma.ValidationError(errors)
 
 
 def _embedded_document_io_validate(field, value):
+    if not value:
+        return
     return _io_validate_data_proxy(value.schema, value._data)
 
 
@@ -296,10 +337,10 @@ class TxMongoReference(Reference):
     def fetch(self, no_data=False, force_reload=False):
         if not self._document or force_reload:
             if self.pk is None:
-                raise ReferenceError('Cannot retrieve a None Reference')
+                raise NoneReferenceError('Cannot retrieve a None Reference')
             self._document = yield self.document_cls.find_one(self.pk)
             if not self._document:
-                raise ValidationError(self.error_messages['not_found'].format(
+                raise ma.ValidationError(self.error_messages['not_found'].format(
                     document=self.document_cls.__name__))
         returnValue(self._document)
 
@@ -308,10 +349,6 @@ class TxMongoBuilder(BaseBuilder):
 
     BASE_DOCUMENT_CLS = TxMongoDocument
 
-    @staticmethod
-    def is_compatible_with(db):
-        return isinstance(db, Database)
-
     def _patch_field(self, field):
         super()._patch_field(field)
 
@@ -326,8 +363,48 @@ class TxMongoBuilder(BaseBuilder):
             field.io_validate = validators
         if isinstance(field, ListField):
             field.io_validate_recursive = _list_io_validate
+        if isinstance(field, DictField):
+            field.io_validate_recursive = _dict_io_validate
         if isinstance(field, ReferenceField):
             field.io_validate.append(_reference_io_validate)
             field.reference_cls = TxMongoReference
         if isinstance(field, EmbeddedField):
             field.io_validate_recursive = _embedded_document_io_validate
+
+
+class TxMongoInstance(Instance):
+    """
+    :class:`umongo.instance.Instance` implementation for txmongo
+    """
+    BUILDER_CLS = TxMongoBuilder
+
+    @staticmethod
+    def is_compatible_with(db):
+        return isinstance(db, Database)
+
+
+class TxMongoMigrationInstance(TxMongoInstance):
+    """TxMongo instance with migration features"""
+
+    @inlineCallbacks
+    def migrate_2_to_3(self):
+        """Migrate database from umongo 2 to umongo 3
+
+        - EmbeddedDocument _cls field is only set if child of concrete embedded document
+        """
+        concrete_not_children = [
+            name for name, ed in self._embedded_lookup.items()
+            if not ed.opts.is_child and not ed.opts.abstract
+        ]
+
+        for doc_cls in self._doc_lookup.values():
+            if doc_cls.opts.abstract:
+                continue
+            if doc_cls.opts.is_child:
+                continue
+            res = yield doc_cls.collection.find()
+            for doc in res:
+                doc = remove_cls_field_from_embedded_docs(doc, concrete_not_children)
+                ret = yield doc_cls.collection.replace_one({"_id": doc["_id"]}, doc)
+                if ret.matched_count != 1:
+                    raise UpdateError(ret)
